using System;
using System.Collections.Generic;
using System.Text;

using System.Data;
using System.Data.Common;
using System.Globalization;

namespace MyKarte.Database
{
    /// <summary>
    /// f[^x[X NX
    /// </summary>
    public abstract class DbBase : IDisposable
    {
        // ڑ
        private DbConnection _conn;

        // f[^A_v^
        private DbDataAdapter _da;

        // gUNV
        private DbTransaction _tran;

        private bool _opened = false;

        private bool _disposed = false;

        #region vpeB
        /// <summary>
        /// ڑ
        /// </summary>
        protected DbConnection Connection
        {
            get
            {
                return _conn;
            }

            set
            {
                _conn = value;
            }
        }
        #endregion

        /// <summary>
        /// DbProviderFactory擾B
        /// </summary>
        /// <returns>DbProviderFactory</returns>
        protected abstract DbProviderFactory GetFactory();

        /// <summary>
        /// DbCommand쐬B
        /// </summary>
        /// <returns>DbCommand</returns>
        public DbCommand CreateCommand()
        {
            DbProviderFactory factory = GetFactory();

            return factory.CreateCommand();
        }

        /// <summary>
        /// DbConnection쐬B
        /// </summary>
        /// <returns>DbConnection</returns>
        protected DbConnection CreateConnection()
        {
            DbProviderFactory factory = GetFactory();

            return factory.CreateConnection();
        }

        /// <summary>
        /// DbDataAdapter쐬B
        /// </summary>
        /// <returns>DbDataAdapter</returns>
        private DbDataAdapter CreateDataAdapter()
        {
            DbProviderFactory factory = GetFactory();

            return factory.CreateDataAdapter();
        }

        /// <summary>
        /// DbCommandBuilder쐬B
        /// </summary>
        /// <returns>DbCommandBuilder</returns>
        private DbCommandBuilder CreateCommandBuilder()
        {
            DbProviderFactory factory = GetFactory();

            return factory.CreateCommandBuilder();
        }

        /// <summary>
        /// ڑ擾B
        /// </summary>
        /// <returns>ڑ</returns>
        protected abstract string GetConnectionString();

        /// <summary>
        /// ڑJB
        /// </summary>
        public virtual void Open()
        {
            _conn.Open();
        }

        /// <summary>
        /// ڑB
        /// </summary>
        public virtual void Close()
        {
            _conn.Close();
        }

        /// <summary>
        /// f[^x[Xf[^擾B
        /// </summary>
        /// <param name="query">SQLNG</param>
        /// <returns>f[^e[u</returns>
        public DataTable GetDataTable(string query)
        {
            DbCommand dbCmd = CreateCommand();
            dbCmd.CommandText = query;

            return GetDataTable(dbCmd);
        }

        /// <summary>
        /// f[^x[Xf[^擾B
        /// </summary>
        /// <param name="dbCmd">DbCommand</param>
        /// <returns>f[^e[u</returns>
        public DataTable GetDataTable(DbCommand dbCmd)
        {
            dbCmd.Connection = _conn;

            _da = CreateDataAdapter();
            _da.SelectCommand = dbCmd;
            _da.SelectCommand.Transaction = _tran;

            DataTable dt = new DataTable();
            dt.Locale = CultureInfo.InvariantCulture;

            // f[^擾
            _da.Fill(dt);

            dbCmd.Parameters.Clear();

            return dt;
        }

        /// <summary>
        /// f[^x[XɃf[^ݒ肷B
        /// </summary>
        /// <param name="dt">f[^e[u</param>
        /// <returns></returns>
        public int SetDataTable(DataTable dt)
        {
            int count = 0;

            bool opened = false;
            bool transacted = false;

            try
            {
                if (_conn.State == ConnectionState.Closed)
                {
                    _conn.Open();

                    opened = true;
                }

                if ((_tran == null) || (_tran.Connection == null))
                {
                    _tran = _conn.BeginTransaction();

                    transacted = true;
                }

                _da.SelectCommand.Transaction = _tran;

                DbCommandBuilder cb = CreateCommandBuilder();
                cb.DataAdapter = _da;

                _da.InsertCommand = cb.GetInsertCommand();
                _da.UpdateCommand = cb.GetUpdateCommand();
                _da.DeleteCommand = cb.GetDeleteCommand();

                _da.InsertCommand.Transaction = _tran;
                _da.UpdateCommand.Transaction = _tran;
                _da.DeleteCommand.Transaction = _tran;

                // XV
                count = _da.Update(dt);

                if (transacted)
                {
                    // R~bg
                    _tran.Commit();
                }
            }
            catch (Exception)
            {
                if (transacted)
                {
                    // [obN
                    _tran.Rollback();
                }

                // OăX[
                throw;
            }
            finally
            {
                if (transacted)
                {
                    _tran.Dispose();
                }

                if (opened)
                {
                    _conn.Close();
                }
            }

            return count;
        }

        /// <summary>
        /// SQLNGsB
        /// </summary>
        /// <param name="query">SQLNG</param>
        /// <returns></returns>
        public int ExecuteSql(string query)
        {
            DbCommand dbCmd = CreateCommand();
            dbCmd.CommandText = query;

            return ExecuteSql(dbCmd);
        }

        /// <summary>
        /// DbCommandsB
        /// </summary>
        /// <param name="dbCmd">DbCommand</param>
        /// <returns></returns>
        public int ExecuteSql(DbCommand dbCmd)
        {
            int count = 0;

            bool opened = false;

            try
            {
                if (_conn.State == ConnectionState.Closed)
                {
                    _conn.Open();

                    opened = true;
                }

                if (_tran != null)
                {
                    dbCmd.Transaction = _tran;
                }

                dbCmd.Connection = _conn;

                // s
                count = dbCmd.ExecuteNonQuery();
            }
            finally
            {
                if (opened)
                {
                    _conn.Close();
                }
            }

            return count;
        }

        /// <summary>
        /// gUNVJnB
        /// </summary>
        public void BeginTransaction()
        {
            if (_conn.State == ConnectionState.Closed)
            {
                _conn.Open();

                _opened = true;
            }

            _tran = _conn.BeginTransaction();
        }

        /// <summary>
        /// R~bgB
        /// </summary>
        public void Commit()
        {
            _tran.Commit();
            _tran.Dispose();

            if (_opened)
            {
                _conn.Close();

                _opened = false;
            }
        }

        /// <summary>
        /// [obNB
        /// </summary>
        public void Rollback()
        {
            _tran.Rollback();
            _tran.Dispose();

            if (_opened)
            {
                _conn.Close();

                _opened = false;
            }
        }

        /// <summary>
        /// e[uꗗ擾B
        /// </summary>
        /// <returns>e[uꗗ</returns>
        public abstract string[] GetTables();

        /// <summary>
        /// e[u݂邩mFB
        /// </summary>
        /// <param name="tableName">e[u</param>
        /// <returns>trueF݂, falseF݂Ȃ</returns>
        public bool ExistTable(string tableName)
        {
            string[] tables = GetTables();

            bool ret = Array.Exists<string>(tables,
                delegate(string param)
                {
                    return param == tableName;
                });

            return ret;
        }

        /// <summary>
        /// fXgN^
        /// </summary>
        ~DbBase()
        {
            Dispose(false);
        }

        /// <summary>
        /// \[XB
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// \[XB
        /// </summary>
        /// <param name="disposing">
        /// }l[W \[XƃA}l[W \[X̗ꍇ trueB
        /// A}l[W \[Xꍇ falseB
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (_disposed)
            {
                return;
            }

            _disposed = true;

            if (disposing)
            {
                _conn.Dispose();

                if (_da != null)
                {
                    _da.Dispose();
                }
            }
        }
    }
}
